import logging
import os
import subprocess
import sys

import pytest
from click.testing import CliRunner

from vja.cli import cli

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)


@pytest.fixture(name='runner', scope='session')
def setup_runner():
    return CliRunner()


def invoke(runner, command, return_code=0, user_input=None, catch_exceptions=False):
    if isinstance(command, str):
        command = command.split()
    res = runner.invoke(cli, command, input=user_input, catch_exceptions=catch_exceptions)
    sys.stdout.write(res.output)
    if res.stderr_bytes:
        sys.stdout.write(res.stderr_bytes)
    if res.exception:
        logging.warning(res.exception)
    assert res.exit_code == return_code, res
    return res


def _login_as_test_user():
    run_vja('vja logout')
    run_vja('vja --username=test --password=test user show')


def _create_project_and_task():
    run_vja('vja project add test-project')
    run_vja('vja project add child --parent-project=test-project')
    run_vja('vja project add grand-child --parent-project=child')
    run_vja('vja bucket add --project=test-project Second bucket')
    run_vja('vja task add At least one task --force-create --priority=5 --due-date=today --label=my_tag --favorite=True --project-id=test-project')
    run_vja('vja task add Task in subproject --force-create --project-id=grand-child')
    run_vja('vja task add A task without a label --force-create')
    run_vja('vja task ls')
    run_vja('vja task show 1')


def run_vja(command):
    result = subprocess.run(command.split(), capture_output=True, check=False)
    if result.returncode:
        print(f'!!! Non-zero result ({result.returncode}) from command {command}')
        sys.stdout.write(result.stdout.decode('utf-8'))
        sys.stdout.write(result.stderr.decode('utf-8'))
        sys.exit(1)


def pytest_configure():
    if 'VJA_CONFIGDIR' not in os.environ:
        print('!!! Precondition not met. You must set VJA_CONFIGDIR in environment variables !!!')
        sys.exit(1)

    _login_as_test_user()

    _create_project_and_task()



import logging
from typing import Optional, List

from vja import VjaError
from vja.apiclient import ApiClient
from vja.model import Project, User

logger = logging.getLogger(__name__)


class ProjectService:
    def __init__(self, api_client: ApiClient):
        self._api_client = api_client
        self._project_by_id: dict[int, Project] = {}

    def find_all_projects(self) -> List[Project]:
        if not self._project_by_id:
            self._project_by_id = {x['id']: Project.from_json(x, []) for x in self._api_client.get_projects()}
            self.fill_ancestors()
        return self._project_by_id.values()

    def find_project_by_id(self, project_id: int) -> Optional[Project]:
        if not self._project_by_id:
            self._project_by_id = {x['id']: Project.from_json(x, []) for x in self._api_client.get_projects()}
            self.fill_ancestors()
        return self._project_by_id.get(project_id)

    def find_project_by_title(self, title) -> Project:
        project_objects = [Project.from_json(x, []) for x in self._api_client.get_projects()]
        if not project_objects:
            raise VjaError('No projects exist. Go and create at least one.')
        project_found = [x for x in project_objects if x.title == title]
        if not project_found:
            raise VjaError(f'Project with title {title} does not exist.')
        return project_found[0]

    def get_default_project(self) -> Project:
        user = User.from_json(self._api_client.get_user())
        project_found = self.find_project_by_id(user.default_project_id)
        if project_found is None:
            project_objects = [Project.from_json(x, []) for x in self._api_client.get_projects()]
            if not project_objects:
                raise VjaError('No projects exist. Go and create at least one.')
            project_objects.sort(key=lambda x: x.id)
            favorite_projects = [x for x in project_objects if x.is_favorite]
            if favorite_projects:
                project_found = favorite_projects[0]
            else:
                project_found = project_objects[0]
        return project_found

    def fill_ancestors(self):
        for project in self._project_by_id.values():
            ancestor_projects = []
            ancestor = self.get_ancestor_project(project.id, project.parent_project_id)
            while ancestor:
                ancestor_projects.append(ancestor)
                ancestor = self.get_ancestor_project(ancestor.id, ancestor.parent_project_id)
            project.ancestor_projects = ancestor_projects

    def get_ancestor_project(self, project_id, parent_project_id) -> Optional[Project]:
        if project_id == parent_project_id or parent_project_id == 0 or project_id == 0:
            return None
        return self._project_by_id.get(parent_project_id)



import logging

from vja.model import Task, Label
from vja.project_service import ProjectService
from vja.urgency import Urgency

logger = logging.getLogger(__name__)


class TaskService:
    def __init__(self, project_service: ProjectService, urgency: Urgency):
        self._project_service = project_service
        self._urgency = urgency

    def task_from_json(self, task_json: dict) -> Task:
        project_object = self._project_service.find_project_by_id(task_json['project_id'])
        labels = Label.from_json_array(task_json['labels'])
        task = Task.from_json(task_json, project_object, labels)
        task.urgency = self._urgency.compute_for(task)
        return task



from datetime import datetime

from vja.config import VjaConfiguration
from vja.model import Task


class Urgency:

    def __init__(self, urgency_coefficients: dict, project_keywords, label_keywords):
        self._urgency_coefficients = urgency_coefficients
        self._project_keywords = project_keywords
        self._label_keywords = label_keywords

    def compute_for(self, task: Task):
        if task.done:
            return 0
        due_date_score = self._get_due_date_score(task) * self._urgency_coefficients.get('due_date_weight', 1.0)
        priority_score = task.priority * self._urgency_coefficients.get('priority_weight', 1.0)
        favorite_score = int(task.is_favorite) * self._urgency_coefficients.get('favorite_weight', 1.0)
        project_name_score = self._get_project_score(task) * self._urgency_coefficients.get('project_keyword', 1.0)
        lable_name_score = self._get_label_score(task) * self._urgency_coefficients.get('label_keyword', 1.0)

        return 1 + due_date_score + priority_score + favorite_score + project_name_score + lable_name_score

    def _get_label_score(self, task):
        task_label_title = task.labels.lower()
        return int(any(label_name.lower() in task_label_title for label_name in
                       self._label_keywords)) if self._label_keywords else 0

    def _get_project_score(self, task):
        task_project_title = task.project.title.lower()
        return int(any(project_name.lower() in task_project_title for project_name in
                       self._project_keywords)) if self._project_keywords else 0

    @staticmethod
    def _get_due_date_score(task: Task):
        if task.due_date:
            due_days = (task.due_date - datetime.today()).days
            if due_days < 0:
                result = 6
            elif due_days == 0:
                result = 5
            elif due_days == 1:
                result = 4
            elif 1 < due_days <= 2:
                result = 3
            elif 2 < due_days <= 5:
                result = 2
            elif 5 < due_days <= 10:
                result = 1
            elif due_days > 14:
                result = -1
            else:
                result = 0
        else:
            result = 0
        return result

    @classmethod
    def from_config(cls, config: VjaConfiguration):
        config_project_keywords = config.get_urgency_project_keywords()
        project_keywords = [x.strip() for x in config_project_keywords.split(',')] if config_project_keywords else []
        config_label_keywords = config.get_urgency_label_keywords()
        label_keywords = [x.strip() for x in config_label_keywords.split(',')] if config_label_keywords else []

        urgency_coefficients = {k: float(v) for k, v in config.get_urgency_coefficients().items()}
        return cls(urgency_coefficients, project_keywords, label_keywords)



import json
import logging
import os

import click
import requests

from vja import VjaError

logger = logging.getLogger(__name__)


class Login:
    def __init__(self, api_url, token_file):
        self._api_url = api_url
        self._token_file = token_file
        self._token = {'access': None}

    @property
    def _access_token(self):
        if not self._token['access']:
            raise KeyError('access token not set! call authenticate()')
        return self._token['access']

    def _load_access_token(self):
        try:
            with open(self._token_file, encoding='utf-8') as token_file:
                data = json.load(token_file)
        except IOError:
            return False
        self._token['access'] = data['token']
        if not self._token['access']:
            return False
        return True

    def _store_access_token(self):
        data = {'token': self._access_token}
        with open(self._token_file, 'w', encoding='utf-8') as token_file:
            json.dump(data, token_file)

    def _create_url(self, path):
        return self._api_url + path

    def validate_access_token(self, force=False, username=None, password=None, totp_passcode=None):
        if self._load_access_token() and not force:
            return
        username = username or click.prompt('Username')
        password = password or click.prompt('Password', hide_input=True)
        response = self._post_login_request(username, password, totp_passcode)
        if response.status_code == 412 and self._to_json(response)['message'] == 'Invalid totp passcode.':
            totp_passcode = totp_passcode or click.prompt('One-time password')
            response = self._post_login_request(username, password, totp_passcode)
        response.raise_for_status()
        self._token['access'] = self._to_json(response)['token']
        self._store_access_token()
        logger.info('Login successful.')

    def get_auth_header(self):
        return {'Authorization': f'Bearer {self._access_token}'}

    def logout(self):
        if os.path.isfile(self._token_file):
            os.remove(self._token_file)

    def _post_login_request(self, username, password, totp_passcode):
        login_url = self._create_url('/login')
        payload = {'long_token': True,
                   'username': username,
                   'password': password,
                   'totp_passcode': totp_passcode}
        return requests.post(login_url, json=payload, timeout=30)

    @staticmethod
    def _to_json(response: requests.Response):
        try:
            return response.json()
        except Exception as e:
            logger.error('Expected valid json, but found %s', response.text)
            raise VjaError('Cannot parse json in response.') from e



import json
import re

from tests.conftest import invoke


class TestUser:
    def test_user_show(self, runner):
        res = invoke(runner, 'user show')
        assert re.search(r'username=\'test\'', res.output)

    def test_user_show_json(self, runner):
        res = invoke(runner, 'user show --json')
        assert json.loads(res.output)['username'] == 'test'

    def test_user_show_jsonvja(self, runner):
        res = invoke(runner, 'user show --jsonvja')
        assert json.loads(res.output)['username'] == 'test'


class TestProject:
    def test_project_ls(self, runner):
        res = invoke(runner, 'project ls')
        assert re.search(r'test-project', res.output)

    def test_project_show(self, runner):
        res = invoke(runner, 'project show 1')
        assert len(res.output) > 0

    def test_project_ls_custom_format(self, runner):
        res = invoke(runner, 'project ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^-?\d*$', line)


class TestBucket:
    def test_bucket_ls(self, runner):
        res = invoke(runner, 'bucket ls --project-id=1')
        assert re.search(r'Backlog', res.output)

    def test_bucket_ls_custom_format(self, runner):
        res = invoke(runner, 'bucket ls --project-id=1 --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)


class TestLabel:
    def test_label_ls(self, runner):
        res = invoke(runner, 'label ls')
        assert re.search(r'my_tag', res.output)

    def test_label_ls_custom_format(self, runner):
        res = invoke(runner, 'label ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)


class TestSingleTask:
    def test_task_show(self, runner):
        res = invoke(runner, 'show 1')
        assert re.search(r'id: 1', res.output)

    def test_task_show_json(self, runner):
        res = invoke(runner, 'show 1 --json')
        assert json.loads(res.output)['id'] == 1

    def test_task_show_jsonvja(self, runner):
        res = invoke(runner, 'show 1 --jsonvja')
        data = json.loads(res.output)
        assert data['id'] == 1
        assert data['title'] is not None
        assert data['project']['id'] is not None
        assert data['created'] is not None
        assert data['updated'] is not None


class TestTaskLs:
    def test_task_ls(self, runner):
        res = invoke(runner, 'ls')
        assert re.search(r'At least one task', res.output)

    def test_task_ls_json(self, runner):
        res = invoke(runner, 'ls --json')
        assert json.loads(res.output)[0]['title'] is not None

    def test_task_ls_jsonvja(self, runner):
        res = invoke(runner, 'ls --jsonvja')
        data = json.loads(res.output)[0]
        assert data['title'] is not None

    def test_filter_id(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '2'])
        data = json.loads(res.output)
        assert data[0]['id'] == 2
        res = invoke(runner, ['ls', '--jsonvja', '1', '2'])
        data = json.loads(res.output)
        assert len(data) == 2
        assert data[0]['id'] in [1, 2]
        assert data[1]['id'] in [1, 2]

    def test_sort_id(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--sort=id'])
        data = json.loads(res.output)
        assert data[0]['id'] == 1
        res = invoke(runner, ['ls', '--jsonvja', '--sort=-id'])
        data = json.loads(res.output)
        assert data[0]['id'] > 1

    def test_sort_combined(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--sort=due_date, is_favorite, -priority, project.title'])
        data = json.loads(res.output)
        assert data[0]['due_date'] is not None
        assert data[-1]['due_date'] is None

    def test_task_custom_format(self, runner):
        res = invoke(runner, 'ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)

    def test_task_custom_format_long(self, runner):
        res = invoke(runner, 'ls --custom-format=tasklist_long')
        assert re.search(r'At least one task', res.output)


class TestTaskLsFilter:
    def test_task_filter_due(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--due-date=after yesterday'])
        data = json.loads(res.output)
        assert len(data) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--due-date=before tomorrow'])
        data = json.loads(res.output)
        assert len(data) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--due-date=after next week'])
        data = json.loads(res.output)
        assert len(data) == 0
        res = invoke(runner, ['ls', '--jsonvja', '--due-date='''''''])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['due_date'] is None for i in data)

    def test_task_filter_favorite(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--favorite=True'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['is_favorite'] for i in data)

    def test_task_filter_label(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--label=y_ta'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert data[0]['label_objects'][0]['title'] == 'my_tag'
        res = invoke(runner, ['ls', '--jsonvja', '--label=unknown_label'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_label_empty(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--label='''''''])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(len(i['label_objects']) == 0 for i in data)

    def test_task_filter_project(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--project=est-project'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['project']['title'] == 'test-project' for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--project=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_base_project(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--base-project=est-project'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['project']['title'] == 'test-project' or i['project']['title'] == 'grand-child' for i in data)
        assert any(i['project']['title'] == 'test-project' for i in data)
        assert any(i['project']['title'] == 'grand-child' for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--project=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_priority(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--priority=eq 5'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['priority'] == 5 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--priority=gt 4'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['priority'] > 4 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--priority=gt 5'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_title(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--title=at least'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all('At least one task' in i['title'] for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--title=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_urgency(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--urgency=10'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['urgency'] >= 10 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--urgency=10000'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_general(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--filter=due_date after 2 days ago'])
        assert len(json.loads(res.output)) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--filter=due_date after 200 days'])
        assert len(json.loads(res.output)) == 0
        res = invoke(runner, ['ls', '--jsonvja', '--filter=due_date after 200 days'])
        assert len(json.loads(res.output)) == 0
        res = invoke(runner, ['ls', '--jsonvja', '--filter=is_favorite eq True'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['is_favorite'] for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--filter=priority gt 2'])
        assert len(json.loads(res.output)) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--filter=priority gt 5'])
        assert len(json.loads(res.output)) == 0
        res = invoke(runner, ['ls', '--jsonvja', '--filter=title contains At least one'])
        assert len(json.loads(res.output)) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--filter=title contains TASK_NOT_CREATED'])
        assert len(json.loads(res.output)) == 0

    def test_task_filter_general_label(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--filter=labels contains my_tag'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all('my_tag' in _labels_from_task_json(task) for task in data)
        res = invoke(runner, ['ls', '--jsonvja', '--filter=labels ne my_tag'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all('my_tag' not in _labels_from_task_json(task) for task in data)

    def test_task_filter_general_combined(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--filter=id gt 0', '--filter=id lt 2'])
        data = json.loads(res.output)
        assert len(data) == 1
        assert all(i['id'] == 1 for i in data)


def _labels_from_task_json(task):
    return ' '.join(map(lambda label: label['title'], task['label_objects'] or []))



import logging

from vja.apiclient import ApiClient
from vja.filter import create_filters
from vja.model import Label, User, Bucket
from vja.parse import rgetattr
from vja.project_service import ProjectService
from vja.task_service import TaskService

logger = logging.getLogger(__name__)
DEFAULT_SORT_STRING = 'done, -urgency, due_date, -priority, project.title, title'


class QueryService:
    def __init__(self, project_service: ProjectService, task_service: TaskService, api_client: ApiClient):
        self._project_service = project_service
        self._task_service = task_service
        self._api_client = api_client

    # user
    def find_current_user(self):
        return User.from_json(self._api_client.get_user())

    # project
    def find_all_projects(self):
        return self._project_service.find_all_projects()

    def find_project_by_id(self, project_id):
        return self._project_service.find_project_by_id(project_id)

    # bucket
    def find_buckets_in_first_kanban_view(self, project_id):
        project = self._project_service.find_project_by_id(project_id)
        project_view = project.get_first_kanban_project_view()
        return Bucket.from_json_array(self._api_client.get_buckets(project_id, project_view.id))

    # label
    def find_all_labels(self):
        return Label.from_json_array(self._api_client.get_labels())

    # tasks
    def find_filtered_tasks(self, include_completed, sort_string, filter_args):
        task_object_array = [self._task_service.task_from_json(x) for x in
                             self._api_client.get_tasks(exclude_completed=not include_completed)]
        filtered_tasks = self._filter(task_object_array, filter_args)
        return self._sort(filtered_tasks, sort_string)

    def find_task_by_id(self, task_id: int):
        return self._task_service.task_from_json(self._api_client.get_task(task_id))

    @staticmethod
    def _filter(task_object_array, filter_args):
        filters = create_filters(filter_args)
        return list(filter(lambda x: all(f(x) for f in filters), task_object_array))

    @staticmethod
    def _sort(filtered_tasks, sort_string):
        sort_string = sort_string or DEFAULT_SORT_STRING
        sort_fields = [{'name': x.strip().strip('-'),
                        'reverse': x.strip().startswith('-')}
                       for x in sort_string.split(',')]
        for sort_field in reversed(sort_fields):
            filtered_tasks.sort(
                key=lambda x, field=sort_field['name']: (
                    sortable_task_value(x, field) is None, sortable_task_value(x, field)),
                reverse=sort_field['reverse'])
        return filtered_tasks


def sortable_task_value(task, field):
    field_name = field
    if field in ('label', 'labels'):
        field_name = 'labels'
    field_value = rgetattr(task, field_name)
    if isinstance(field_value, str):
        return field_value.upper()
    return field_value



import logging
import sys

import requests

from vja import VjaError
from vja.authenticate import Login

logger = logging.getLogger(__name__)


def inject_access_token(func):
    def wrapper(self, *args, **kwargs):
        try:
            headers = self.authenticate(force_login=False)
            return func(self, headers=headers, *args, **kwargs)
        except KeyError as e:
            raise VjaError(f'need access token to call function {func.__name__}; call authenticate()') from e

    return wrapper


def handle_http_error(func):
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except requests.HTTPError as error:
            if error.response.status_code == 401:
                logger.info('HTTP-Error %s, url=%s; trying to retrieve new access token...',
                            error.response.status_code, error.response.url)
                self.authenticate(force_login=True)
                return func(self, *args, **kwargs)
            logger.warning('HTTP-Error %s, url=%s, body=%s',
                           error.response.status_code, error.response.url, error.response.text)
            sys.exit(1)

    return wrapper


class ApiClient:
    def __init__(self, api_url, token_file):
        logger.debug('Connecting to api_url %s', api_url)
        self._api_url = api_url
        self._cache = {'projects': None, 'labels': None, 'tasks': None}
        self._login = Login(api_url, token_file)

    def _create_url(self, path):
        return self._api_url + path

    @handle_http_error
    @inject_access_token
    def _get_json(self, url, params=None, headers=None):
        if params is None:
            params = {}
        response = requests.get(url, headers=headers, params=params, timeout=30)
        # to verbose:
        # logger.debug("GET response: %s - %s", response, response.text)
        response.raise_for_status()
        json_result = self._to_json(response)
        total_pages = int(response.headers.get('x-pagination-total-pages', 1))
        if total_pages > 1:
            logger.debug('Trying to load all pages. Consider to increase MAXITEMSPERPAGE on your server instead.')
            for page in range(2, total_pages + 1):
                logger.debug('load page %s', page)
                params.update({'page': page})
                response = requests.get(url, headers=headers, params=params, timeout=30)
                response.raise_for_status()
                json_result = json_result + self._to_json(response)
        return json_result

    @handle_http_error
    @inject_access_token
    def _put_json(self, url, params=None, payload=None, headers=None):
        response = requests.put(url, headers=headers, params=params, json=payload, timeout=30)
        logger.debug("PUT response: %s - %s", response, response.text)
        response.raise_for_status()
        return self._to_json(response)

    @handle_http_error
    @inject_access_token
    def _post_json(self, url, params=None, payload=None, headers=None):
        response = requests.post(url, headers=headers, params=params, json=payload, timeout=30)
        logger.debug("POST response: %s - %s", response, response.text)
        response.raise_for_status()
        return self._to_json(response)

    @handle_http_error
    @inject_access_token
    def _delete_json(self, url, params=None, payload=None, headers=None):
        response = requests.delete(url, headers=headers, params=params, json=payload, timeout=30)
        logger.debug("DELETE response: %s - %s", response, response.text)
        response.raise_for_status()
        return self._to_json(response)

    @staticmethod
    def _to_json(response: requests.Response):
        try:
            return response.json()
        except Exception as e:
            logger.error('Expected valid json, but found %s', response.text)
            raise VjaError('Cannot parse json in response.') from e

    def authenticate(self, force_login=True, username=None, password=None, totp_passcode=None):
        try:
            self._login.validate_access_token(force_login, username, password, totp_passcode)
            return self._login.get_auth_header()
        except requests.HTTPError as error:
            logger.warning('HTTP-Error %s, url=%s, body=%s',
                           error.response.status_code, error.response.url, error.response.text)
            sys.exit(1)

    def logout(self):
        self._login.logout()

    def get_user(self):
        return self._get_json(self._create_url('/user'))

    def get_projects(self):
        if self._cache['projects'] is None:
            self._cache['projects'] = self._get_json(self._create_url('/projects')) or []
        return self._cache['projects']

    def get_project(self, project_id):
        return self._get_json(self._create_url(f'/projects/{str(project_id)}'))

    def put_project(self, parent_project_id, title):
        payload = {'title': title, 'parent_project_id': parent_project_id}
        return self._put_json(self._create_url('/projects'), payload=payload)

    def get_buckets(self, project_id, project_view_id):
        return self._get_json(self._create_url(f'/projects/{str(project_id)}/views/{str(project_view_id)}/tasks'))

    def put_bucket(self, project_id, project_view_id, title):
        payload = {'title': title}
        return self._put_json(self._create_url(f'/projects/{str(project_id)}/views/{str(project_view_id)}/buckets'),
                              payload=payload)

    def get_labels(self):
        if self._cache['labels'] is None:
            self._cache['labels'] = self._get_json(self._create_url('/labels')) or []
        return self._cache['labels']

    def put_label(self, title):
        payload = {'title': title}
        return self._put_json(self._create_url('/labels'), payload=payload)

    def get_tasks(self, exclude_completed=True):
        if self._cache['tasks'] is None:
            url = self._create_url('/tasks/all')
            params = {'filter': 'done=false'} if exclude_completed else {}
            self._cache['tasks'] = self._get_json(url, params) or []
        return self._cache['tasks']

    def get_task(self, task_id):
        url = self._create_url(f'/tasks/{str(task_id)}')
        return self._get_json(url)

    def put_task(self, project_id, payload):
        return self._put_json(self._create_url(f'/projects/{str(project_id)}/tasks'), payload=payload)

    def post_task(self, task_id, payload):
        return self._post_json(self._create_url(f'/tasks/{str(task_id)}'), payload=payload)

    def add_label_to_task(self, task_id, label_id):
        task_label_url = self._create_url(f'/tasks/{str(task_id)}/labels')
        payload = {'label_id': label_id}
        return self._put_json(task_label_url, payload=payload)

    def remove_label_from_task(self, task_id, label_id):
        task_label_url = self._create_url(f'/tasks/{str(task_id)}/labels/{str(label_id)}')
        self._delete_json(task_label_url)



import functools
import re
import time
from datetime import datetime, timedelta
from html.parser import HTMLParser
from typing import Optional

from dateutil import tz, parser
from parsedatetime import parsedatetime

# if dates are given like 'tomorrow' or 'next mon' then this will be used as time
DEFAULT_DATE_HOUR = 8
DEFAULT_DATE_MINUTE = 0

_timedelta_regex = re.compile(r'^((?P<weeks>[.\d]+?)w)? *'
                              r'^((?P<days>[.\d]+?)d)? *'
                              r'((?P<hours>[.\d]+?)h)? *'
                              r'((?P<minutes>[.\d]+?)m)? *'
                              r'((?P<seconds>[.\d]+?)s?)?$')


def parse_date_arg_to_datetime(text: str,
                               default_hour=DEFAULT_DATE_HOUR,
                               default_minute=DEFAULT_DATE_MINUTE) -> Optional[datetime]:
    if not text:
        return None

    if re.compile(r'.*\dT\d.*').match(text):
        text = text.replace("T", " ")  # workaround for https://github.com/bear/parsedatetime/issues/15

    timetuple, pdt_context = parsedatetime.Calendar(version=parsedatetime.VERSION_CONTEXT_STYLE).parse(text)
    result = datetime.fromtimestamp(time.mktime(timetuple))
    if not pdt_context.hasTime:
        result = result.replace(hour=default_hour, minute=default_minute, second=0)
    return result


def parse_date_arg_to_iso(text: str,
                          default_hour=DEFAULT_DATE_HOUR,
                          default_minute=DEFAULT_DATE_MINUTE) -> Optional[str]:
    date_value = parse_date_arg_to_datetime(text, default_hour, default_minute)
    return datetime_to_isoformat(date_value) if date_value else None


def datetime_to_isoformat(date: datetime) -> Optional[str]:
    return date.astimezone(tz.tzlocal()).isoformat() if date else None


def parse_date_arg_to_timedelta(time_str: str) -> Optional[timedelta]:
    """
    Parse a time string e.g. '2h 13m' or '1.5d' into a timedelta object.
    Based on Peter's answer at https://stackoverflow.com/a/51916936/2445204
    and virhilo's answer at https://stackoverflow.com/a/4628148/851699
    :param time_str: A string identifying a duration, e.g. '2h13.5m'
    :return datetime.timedelta: A datetime.timedelta object
    """
    if not time_str:
        return None
    parts = _timedelta_regex.match(time_str)
    assert parts is not None, f"""Could not parse any time information from '{time_str}'.
    Examples of valid strings: '8h', '2d 8h 5m 2s', '2m4.3s'"""
    time_params = {name: float(param)
                   for name, param in parts.groupdict().items() if param}
    return timedelta(**time_params)


def parse_json_date(json_date: str) -> Optional[datetime]:
    if json_date and json_date > '0001-01-02T00:00:00Z':
        return parser.isoparse(json_date).astimezone(tz.tzlocal()).replace(tzinfo=None)
    return None


def parse_bool_arg(text: str) -> bool:
    return text.lower() in ['true', '1', 't', 'y', 'yes'] if text else False


class HTMLFilter(HTMLParser):
    inner_text = ""

    def handle_data(self, data):
        self.inner_text += data


def html2text(text: str) -> str:
    parser = HTMLFilter()
    parser.feed(text)
    return parser.inner_text


def rgetattr(obj, path: str, *default):
    attrs = path.split('.')
    try:
        return functools.reduce(getattr, attrs, obj)
    except AttributeError:
        if default:
            return default[0]
        raise



import datetime
import logging

from vja import VjaError
from vja.apiclient import ApiClient
from vja.model import Bucket, Label, Project
from vja.parse import parse_date_arg_to_iso, parse_json_date, parse_date_arg_to_timedelta, datetime_to_isoformat
from vja.project_service import ProjectService
from vja.task_service import TaskService

logger = logging.getLogger(__name__)


class CommandService:
    def __init__(self, project_service: ProjectService, task_service: TaskService, api_client: ApiClient):
        self._project_service = project_service
        self._task_service = task_service
        self._api_client = api_client

    def login(self, username, password, totp_passcode):
        self._api_client.authenticate(True, username, password, totp_passcode)

    def logout(self):
        self._api_client.logout()
        logger.info('Logged out')

    # project
    def add_project(self, parent_project, title):
        if parent_project:
            if str(parent_project).isdigit():
                parent_project_id = parent_project
            else:
                parent_project_id = self._project_service.find_project_by_title(parent_project).id
        else:
            parent_project_id = None
        project_json = self._api_client.put_project(parent_project_id, title)
        return Project.from_json(project_json, [])

    # bucket
    def add_bucket(self, project_query, title):
        if str(project_query).isdigit():
            project = self._project_service.find_project_by_id(project_query)
        else:
            project = self._project_service.find_project_by_title(project_query)
        project_view = next(x for x in project.views if x.view_kind == "kanban")
        bucket_json = self._api_client.put_bucket(project.id, project_view.id, title)
        return Bucket.from_json(bucket_json)

    # label
    def add_label(self, title):
        label_json = self._api_client.put_label(title)
        return Label.from_json(label_json)

    # tasks
    _arg_to_json = {'title': {'field': 'title', 'mapping': (lambda x: x)},
                    'note': {'field': 'description', 'mapping': (lambda x: x)},
                    'prio': {'field': 'priority', 'mapping': int},
                    'due': {'field': 'due_date', 'mapping': (lambda x: x)},
                    'favorite': {'field': 'is_favorite', 'mapping': bool},
                    'completed': {'field': 'done', 'mapping': bool},
                    'position': {'field': 'position', 'mapping': int},
                    'project_id': {'field': 'project_id', 'mapping': int},
                    'bucket_id': {'field': 'bucket_id', 'mapping': int},
                    'kanban_position': {'field': 'kanban_position', 'mapping': int},
                    'reminder': {'field': 'reminders', 'mapping': (lambda x: x)}
                    }

    def _args_to_payload(self, args: dict):
        payload = {}
        for arg_name, arg_value in args.items():
            mapper = self._arg_to_json[arg_name]
            payload[mapper['field']] = mapper['mapping'](arg_value)
        return payload

    def add_task(self, title, args: dict):
        args.update({'title': title})
        if args.get('project_id'):
            project_arg = args.pop('project_id')
            if str(project_arg).isdigit():
                project_id = project_arg
            else:
                project_id = self._project_service.find_project_by_title(project_arg).id
        else:
            project_id = self._project_service.get_default_project().id
        label_names = args.pop('label')
        is_force = args.pop('force_create') if args.get('force_create') is not None else False
        if args.get('due') is not None:
            args.update({'due': parse_date_arg_to_iso(args.get('due'))})
        self._parse_reminder_arg(args.get('reminder'), args)

        payload = self._args_to_payload(args)

        if not is_force:
            self._validate_add_task(title, label_names)
        logger.debug('put task: %s', payload)
        task_json = self._api_client.put_task(project_id, payload)
        task = self._task_service.task_from_json(task_json)
        for label_name in label_names:
            label = self._label_from_name(label_name, is_force) if label_name else None
            if label:
                self._api_client.add_label_to_task(task.id, label.id)
        return task

    def clone_task(self, task_id: int, title: str):
        task_remote = self._api_client.get_task(task_id)
        task_remote.update({'id': None})
        task_remote.update({'title': title})
        task_remote.update({'position': 0})
        task_remote.update({'bucket_id': 0})

        logger.debug('put task: %s', task_remote)
        task_json = self._api_client.put_task(task_remote['project_id'], task_remote)
        task = self._task_service.task_from_json(task_json)

        for label in task_remote['labels']:
            self._api_client.add_label_to_task(task.id, label['id'])
        return task

    def edit_task(self, task_id: int, args: dict):
        task_remote = self._api_client.get_task(task_id)
        label_name = args.pop('label') if args.get('label') else None
        is_force = args.pop('force_create') if args.get('force_create') is not None else False

        if args.get('due') is not None:
            self._update_due_date(args, task_remote)
        self._update_reminder(args, task_remote)
        if args.get('note_append'):
            append_note = args.pop('note_append')
            args.update({
                'note': task_remote['description'] + '\n' + append_note if task_remote['description'] else append_note
            })

        payload = self._args_to_payload(args)
        logger.debug('update fields: %s', payload)
        task_remote.update(payload)
        logger.debug('post task: %s', task_remote)
        task_json = self._api_client.post_task(task_id, task_remote)
        task_new = self._task_service.task_from_json(task_json)

        label = self._label_from_name(label_name, is_force) if label_name else None
        if label:
            if task_new.has_label(label):
                self._api_client.remove_label_from_task(task_new.id, label.id)
            else:
                self._api_client.add_label_to_task(task_new.id, label.id)
        return task_new

    @staticmethod
    def _update_due_date(args, task_remote):
        # keep using time of remote task, if none is given
        arg_due = args.get('due')
        remote_date = parse_json_date(task_remote['due_date'])
        if remote_date:
            arg_date = parse_date_arg_to_iso(arg_due, remote_date.hour, remote_date.minute)
        else:
            arg_date = parse_date_arg_to_iso(arg_due)
        args.update({'due': arg_date})

    @staticmethod
    def _update_reminder(args, task_remote):
        reminder_arg = args.get('reminder')
        CommandService._parse_reminder_arg(reminder_arg, args)

        # replace the first existing reminder with our entry
        new_reminder = args.pop('reminder')[0] if reminder_arg else None
        if new_reminder is not None:
            old_reminders = task_remote['reminders']
            if old_reminders and len(old_reminders) > 0:
                if new_reminder:
                    old_reminders[0] = new_reminder  # overwrite first remote reminder
                else:
                    old_reminders.pop(0)  # remove first remote reminder
            else:
                if new_reminder:
                    old_reminders = [new_reminder]  # create single reminder
            args.update({'reminder': old_reminders})

    @staticmethod
    def _parse_reminder_arg(reminder_arg, args):
        if reminder_arg is None:
            return
        if reminder_arg == 'due':
            args.update(
                {'reminder': [{'relative_to': 'due_date', 'relative_period': 0}]})  # --reminder=due or --reminder
        elif 'due' in reminder_arg:
            reminder_due_args = reminder_arg.split(" ", 2)
            duration = int(parse_date_arg_to_timedelta(reminder_due_args[0]).total_seconds())
            sign = -1 if reminder_due_args[1] == 'before' else 1
            args.update(
                {'reminder': [{'relative_to': 'due_date',
                               'relative_period': sign * duration}]})  # --reminder="1h before due_date"
        elif reminder_arg == '':
            args.update(
                {'reminder': None})  # --reminder=""
        else:
            args.update(
                {'reminder': [{'reminder': parse_date_arg_to_iso(reminder_arg)}]})

    def toggle_task_done(self, task_id):
        task_remote = self._api_client.get_task(task_id)
        task_remote.update({'done': not task_remote['done']})
        task_json = self._api_client.post_task(task_id, task_remote)
        return self._task_service.task_from_json(task_json)

    def defer_task(self, task_id, delay_by):
        timedelta = parse_date_arg_to_timedelta(delay_by)
        args = {}

        task_remote = self._api_client.get_task(task_id)
        due_date = parse_json_date(task_remote['due_date'])
        if due_date:
            now = datetime.datetime.now().replace(microsecond=0)
            if due_date < now:
                args.update({'due': datetime_to_isoformat(now + timedelta)})
            else:
                args.update({'due': datetime_to_isoformat(due_date + timedelta)})

        old_reminders = task_remote['reminders']
        if old_reminders and len(old_reminders) > 0:
            reminder_date = parse_json_date(old_reminders[0]['reminder'])
            is_absolute_reminder = not old_reminders[0]['relative_to']
            if reminder_date and is_absolute_reminder:
                args.update({'reminder': datetime_to_isoformat(reminder_date + timedelta)})
                self._update_reminder(args, task_remote)

        payload = self._args_to_payload(args)
        logger.debug('update fields: %s', payload)
        task_remote.update(payload)
        task_json = self._api_client.post_task(task_id, task_remote)
        return self._task_service.task_from_json(task_json)

    def _label_from_name(self, name, is_force):
        if not name:
            return None
        labels_remote = Label.from_json_array(self._api_client.get_labels())
        label_found = [label for label in labels_remote if label.title == name]
        if not label_found:
            if is_force:
                return Label.from_json(self._api_client.put_label(name))
            logger.warning("Ignoring non existing label [%s]. You may want to execute \"label add\" first.", name)
            return None
        return label_found[0]

    def _validate_add_task(self, title, label_names):
        tasks_remote = self._api_client.get_tasks(exclude_completed=True)
        if any(task for task in tasks_remote if task['title'] == title):
            raise VjaError("Task with title does exist. You may want to run with --force-create.")
        for label_name in label_names:
            labels_remote = Label.from_json_array(self._api_client.get_labels())
            if not any(label for label in labels_remote if label.title == label_name):
                raise VjaError(
                    "Label does not exist. You may want to execute \"label add\" or run with --force-create.")



import datetime
import json
import re

from tests.conftest import invoke

ADD_SUCCESS_PATTERN = re.compile(r'.*Created task (\d+) in project .*')
TODAY = datetime.datetime.now().replace(microsecond=0)
TODAY_ISO = TODAY.isoformat()
YESTERDAY = TODAY + datetime.timedelta(days=-1)
YESTERDAY_ISO = YESTERDAY.isoformat()
TOMORROW = TODAY + datetime.timedelta(days=1)
TOMORROW_ISO = TOMORROW.isoformat()
TOMMORROW_AT_8_ISO = (TOMORROW.replace(hour=8, minute=0, second=0)).isoformat()
DATE_1 = TODAY + datetime.timedelta(days=10)
DATE_2 = DATE_1 + datetime.timedelta(days=1)
DATE_1_ISO = DATE_1.isoformat()
DATE_2_ISO = DATE_2.isoformat()


class TestAddTask:
    def test_project_id(self, runner):
        res = invoke(runner, 'add title of new task --force --project=1')
        after = json_for_created_task(runner, res.output)
        assert after['project']['id'] == 1

    def test_project_title(self, runner):
        res = invoke(runner, 'add title of new task --force --project=test-project')
        after = json_for_created_task(runner, res.output)
        assert after['project']['title'] == 'test-project'

    def test_due_date(self, runner):
        res = invoke(runner, 'add title of new task --force --due=tomorrow')
        after = json_for_created_task(runner, res.output)
        assert after['due_date'] == TOMMORROW_AT_8_ISO

    def test_duplicate_task_title_rejected(self, runner):
        invoke(runner, 'add title of new task', 1, catch_exceptions=True)

    def test_default_reminder_uses_due(self, runner):
        res = invoke(runner, 'add title of new task --force --project=test-project --due=today --reminder')
        after = json_for_created_task(runner, res.output)
        assert after['reminders'][0]['relative_period'] == 0
        assert after['reminders'][0]['relative_to'] == 'due_date'

    def test_default_reminder_with_absolute_time(self, runner):
        res = invoke(runner, f'add title of new task --force --project=test-project --reminder={DATE_2_ISO}')
        after = json_for_created_task(runner, res.output)
        assert after['reminders'][0]['reminder'] == DATE_2_ISO

    def test_add_with_multiple_labels(self, runner):
        res = invoke(runner, 'add multi labels --force -l tag_1 -l tag_2')
        after = json_for_created_task(runner, res.output)
        assert has_label_with_title(after['label_objects'], 'tag_1')
        assert has_label_with_title(after['label_objects'], 'tag_2')



class TestCloneTask:
    def test_clone_task(self, runner):
        before = json_for_task_id(runner, 1)
        res = invoke(runner, 'clone 1 title of new task cloned from 1')
        after = json_for_created_task(runner, res.output)
        assert after['project'] == before['project']
        assert after['due_date'] == before['due_date']
        assert after['label_objects'] == before['label_objects']
        assert after['title'] != before['title']
        assert after['id'] != before['id']
        assert after['created'] != before['created']


class TestEditGeneral:
    def test_edit_title(self, runner):
        before = json_for_task_id(runner, 1)
        new_title = f'{before["title"]}42'
        invoke(runner, ['edit', '1', '-i', f'{new_title}'])

        after = json_for_task_id(runner, 1)
        assert after['title'] == new_title
        assert after['updated'] >= before['updated']
        # other attributes remain in place
        assert after['due_date'] == before['due_date']
        assert after['reminders'] == before['reminders']
        assert after['position'] == before['position']
        assert after['project']['id'] == before['project']['id']
        assert after['created'] == before['created']

    def test_edit_due_date_with_time(self, runner):
        # unset due_date
        before = json_for_task_id(runner, 1)
        assert before['due_date'] is not None
        invoke(runner, 'edit 1 --due-date=')
        after = json_for_task_id(runner, 1)
        assert after['due_date'] is None

        # edit due_date with default time
        invoke(runner, 'edit 1 --due=tomorrow')
        after = json_for_task_id(runner, 1)
        assert after['due_date'] == TOMMORROW_AT_8_ISO

        # edit due_date with given time
        invoke(runner, ['edit', '1', '--due=tomorrow 15:00'])
        after = json_for_task_id(runner, 1)
        assert after['due_date'] == (TOMORROW.replace(hour=15, minute=0, second=0)).isoformat()

        # edit due_date without time
        invoke(runner, 'edit 1 --due=today')
        after = json_for_task_id(runner, 1)
        assert after['due_date'] == (TODAY.replace(hour=15, minute=0, second=0)).isoformat()


    def test_toggle_label(self, runner):
        labels_0 = json_for_task_id(runner, 1)['label_objects']
        invoke(runner, 'edit 1 --label=tag1 --force-create')
        labels_1 = json_for_task_id(runner, 1)['label_objects']
        invoke(runner, 'edit 1 --label=tag1')
        labels_2 = json_for_task_id(runner, 1)['label_objects']

        assert labels_0 != labels_1
        assert labels_0 == labels_2
        assert has_label_with_title(labels_0, 'tag1') or has_label_with_title(labels_1, 'tag1')

    def test_append_note(self, runner):
        invoke(runner, 'edit 1 --note=line1')
        note_1 = json_for_task_id(runner, 1)['description']
        invoke(runner, 'edit 1 --note-append=line2')
        note_2 = json_for_task_id(runner, 1)['description']

        assert note_1 == 'line1'
        assert note_2 == 'line1\nline2'

    def test_edit_project(self, runner):
        invoke(runner, 'project add another project')
        invoke(runner, 'edit 1 --project-id=1')
        project_1 = json_for_task_id(runner, 1)['project']['id']
        invoke(runner, 'edit 1 -o 2')
        project_2 = json_for_task_id(runner, 1)['project']['id']

        assert project_1 == 1
        assert project_2 == 2


class TestEditReminder:
    def test_set_reminder_to_absolute_value(self, runner):
        invoke(runner, f'edit 2 --reminder={DATE_1_ISO}')
        before = json_for_task_id(runner, 2)

        invoke(runner, f'edit 2 --due={DATE_1_ISO} --reminder={DATE_2_ISO}')

        after = json_for_task_id(runner, 2)
        assert before['reminders'][0]['reminder'] == DATE_1_ISO
        assert after['reminders'][0]['reminder'] == DATE_2_ISO
        assert not after['reminders'][0]['relative_to']

    def test_set_reminder_to_due(self, runner):
        invoke(runner, f'edit 2 --reminder={DATE_1_ISO}')
        before = json_for_task_id(runner, 2)

        invoke(runner, f'edit 2 --due={DATE_2_ISO} --reminder=due')

        after = json_for_task_id(runner, 2)
        assert before['reminders'][0]['reminder'] == DATE_1_ISO
        assert after['reminders'][0]['reminder'] == DATE_2_ISO
        assert after['reminders'][0]['relative_period'] == 0
        assert after['reminders'][0]['relative_to'] == 'due_date'

    def test_set_reminder_to_due_empty_option(self, runner):
        invoke(runner, f'edit 2 --reminder={DATE_1_ISO}')

        invoke(runner, f'edit 2 --due={DATE_2_ISO} --reminder')

        after = json_for_task_id(runner, 2)
        assert after['reminders'][0]['reminder'] == DATE_2_ISO
        assert after['reminders'][0]['relative_period'] == 0
        assert after['reminders'][0]['relative_to'] == 'due_date'

    def test_set_reminder_to_relative_value(self, runner):
        invoke(runner, f'edit 2 --reminder={DATE_1_ISO}')

        invoke(runner, ['edit', '2', '--due={DATE_2_ISO}', '--reminder=1h1m before due_date'])

        after = json_for_task_id(runner, 2)
        assert after['reminders'][0]['relative_period'] == -3660
        assert after['reminders'][0]['relative_to'] == 'due_date'

    def test_unset_reminder(self, runner):
        invoke(runner, f'edit 2 --reminder={DATE_1_ISO}')
        before = json_for_task_id(runner, 2)

        invoke(runner, 'edit 2 --reminder=')

        after = json_for_task_id(runner, 2)
        assert before['reminders'][0]['reminder'] == DATE_1_ISO
        assert not after['reminders']


class TestToggleDoneTask:
    def test_toggle_done(self, runner):
        done_0 = json_for_task_id(runner, 1)['done']
        invoke(runner, 'check 1')
        done_1 = json_for_task_id(runner, 1)['done']
        invoke(runner, 'check 1')
        done_2 = json_for_task_id(runner, 1)['done']
        assert done_0 != done_1
        assert done_0 == done_2

    def test_toggle_does_not_modify_other_fields(self, runner):
        invoke(runner, 'edit 1 --priority=5')
        json_0 = json_for_task_id(runner, 1)
        invoke(runner, 'check 1')
        json_1 = json_for_task_id(runner, 1)
        invoke(runner, 'check 1')
        json_2 = json_for_task_id(runner, 1)
        assert json_0['priority'] == json_1['priority'] == json_2['priority'] == 5


class TestDeferTask:
    def test_defer_due_date_and_reminder(self, runner):
        invoke(runner, f'edit 2 --due-date={DATE_1_ISO} --reminder={DATE_1_ISO}')

        invoke(runner, 'defer 2 1d')

        after = json_for_task_id(runner, 2)
        assert after['due_date'] == DATE_2_ISO
        assert after['reminders'][0]['reminder'] == DATE_2_ISO

    def test_dont_defer_relative_reminder(self, runner):
        invoke(runner, f'edit 2 --due-date={DATE_1_ISO} -r')

        invoke(runner, 'defer 2 1d')

        after = json_for_task_id(runner, 2)
        assert after['due_date'] == DATE_2_ISO
        assert after['reminders'][0]['relative_period'] == 0
        assert after['reminders'][0]['relative_to'] == 'due_date'

    def test_defer_past_due_realtive_to_now(self, runner):
        invoke(runner, f'edit 2 --due-date={YESTERDAY_ISO}')

        invoke(runner, 'defer 2 1d')

        after = json_for_task_id(runner, 2)
        assert after['due_date'][:10] == TOMORROW_ISO[:10]


class TestMultipleTasks:
    def test_edit_three_tasks(self, runner):
        invoke(runner, 'edit 1 2 3 --priority=4')
        for i in range(1, 4):
            after = json_for_task_id(runner, i)
            assert after['priority'] == 4
            assert after['updated'][:10] == TODAY_ISO[:10]
        invoke(runner, 'edit 1 2 3 --priority=5')

    def test_show_three_tasks(self, runner):
        res = invoke(runner, 'show 1 2 3')
        assert re.search(r'id: 1', res.output)
        assert re.search(r'id: 2', res.output)
        assert re.search(r'id: 3', res.output)


def json_for_created_task(runner, message):
    assert re.match(ADD_SUCCESS_PATTERN, message), message
    task_id = ADD_SUCCESS_PATTERN.findall(message)[0]
    return json_for_task_id(runner, task_id)


def json_for_task_id(runner, task_id):
    res = invoke(runner, f'show {task_id} --jsonvja')

    assert res.exit_code == 0, res
    data = json.loads(res.output)
    return data

def has_label_with_title(labels, title):
    label_titles = [x['title'] for x in labels]
    return title in label_titles




#!/usr/bin/env python

import logging
import webbrowser
from importlib import metadata

import click
from click_aliases import ClickAliasedGroup

from vja.apiclient import ApiClient
from vja.config import VjaConfiguration
from vja.output import Output
from vja.project_service import ProjectService
from vja.service_command import CommandService
from vja.service_query import QueryService
from vja.task_service import TaskService
from vja.urgency import Urgency

logger = logging.getLogger(__name__)


class Application:
    def __init__(self):
        self._configuration = VjaConfiguration()
        api_client = ApiClient(self._configuration.get_api_url(), self._configuration.get_token_file())
        project_service = ProjectService(api_client)
        urgency_service = Urgency.from_config(self._configuration)
        task_service = TaskService(project_service, urgency_service)
        self._command_service = CommandService(project_service, task_service, api_client)
        self._query_service = QueryService(project_service, task_service, api_client)
        self._output = Output()

    @property
    def command_service(self) -> CommandService:
        return self._command_service

    @property
    def query_service(self) -> QueryService:
        return self._query_service

    @property
    def output(self) -> Output:
        return self._output

    @property
    def configuration(self) -> VjaConfiguration:
        return self._configuration

    def open_browser_task(self, task_id):
        url = self.configuration.get_frontend_url().rstrip('/')
        if task_id:
            url += f'/tasks/{str(task_id)}'
        webbrowser.open_new_tab(url)

    def open_browser_project(self, project_id):
        url = self.configuration.get_frontend_url().rstrip('/') + f'/projects/{str(project_id)}'
        webbrowser.open_new_tab(url)


with_application = click.make_pass_decorator(Application, ensure=True)


@click.group(cls=ClickAliasedGroup, context_settings=dict({'help_option_names': ['-h', '--help']}))
@click.pass_context
@click.version_option(metadata.version("vja"))
@click.option('-v', '--verbose', 'verbose', default=False, is_flag=True, help='Activate verbose logging')
@click.option('-u', '--username', 'username', help='Username for initial login (optional)')
@click.option('-p', '--password', 'password', help='Password for initial login (optional)')
@click.option('-t', '--totp-passcode', '--totp_passcode', 'totp_passcode',
              help='Time-based one-time passcode from your authenticator app (optional). '
                   'Only if TOTP is enabled on server.')
def cli(ctx=None, verbose=None, username=None, password=None, totp_passcode=None):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
        logger.debug('Verbose mode on')
    else:
        logging.basicConfig(level=logging.INFO)
        logging.getLogger().setLevel(logging.INFO)
    application = Application()
    ctx.obj = application
    if username:
        application.command_service.login(username, password, totp_passcode)


# user
@cli.group('user', help='Subcommand: user (see help)')
def user_group():
    # vja user
    pass


@user_group.command('show', help='Print current user')
@click.option('is_json', '--json', default=False, is_flag=True, help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True, help='Print as vja application json')
@with_application
def user_show(application, is_json=False, is_jsonvja=False):
    application.output.user(
        application.query_service.find_current_user(), is_json, is_jsonvja)


# projects
@cli.group('project', help='Subcommand: project (see help)', aliases=['projects'])
def project_group():
    # vja project
    pass


@project_group.command('add', help='Add project with title')
@click.option('parent_project', '-o', '--parent-project', '--parent_project',
              help='Create project as child of parent project. May be given by id or title of parent-project.')
@click.argument('title', nargs=-1, required=True)
@with_application
def project_add(application, title, parent_project=None):
    project = application.command_service.add_project(parent_project, " ".join(title))
    click.echo(f'Created project {project.id}')


@project_group.command('ls', help='Print projects ... (id; title; description; parent_project_id)')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='Print as vja application json')
@click.option('custom_format', '--custom-format',
              help='Format with template from .vjacli/vja.rc')
@with_application
def project_ls(application, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.project_array(
        application.query_service.find_all_projects(), is_json, is_jsonvja, custom_format)


@project_group.command('show', help='Show project details')
@click.argument('project_id', required=True, type=click.INT)
@click.option('is_json', '--json', default=False, is_flag=True,
              help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='Print as vja application json')
@with_application
def project_show(application, project_id, is_json, is_jsonvja):
    application.output.project(
        application.query_service.find_project_by_id(project_id), is_json, is_jsonvja)


@project_group.command('open', help='Open project in webbrowser')
@click.argument('project_id', required=True, type=click.INT)
@with_application
def project_open(application, project_id):
    application.open_browser_project(project_id)


# buckets
@cli.group('bucket', help='Subcommand: Kanban buckets (see help)', aliases=['buckets'])
def bucket_group():
    # vja bucket
    pass


@bucket_group.command('add',
                      help='Add bucket with title to the first Kanban View of the project')
@click.option('project', '-o', '--project', '--project-id',
              help='Create bucket in given project.')
@click.argument('title', nargs=-1, required=True)
@with_application
def bucket_add(application, title, project):
    bucket = application.command_service.add_bucket(project, " ".join(title))
    click.echo(f'Created bucket {bucket.id} in project {project}')


@bucket_group.command('ls',
                      help='Show Kanban buckets of given project (only the first Kanban View)... '
                           '(id; title; limit; count tasks)')
@click.option('project_id', '-o', '--project', '--project-id', '--project_id', required=True, type=click.INT,
              help='Show buckets of project with id')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='Print as vja application json')
@click.option('custom_format', '--custom-format',
              help='Format with template from .vjacli/vja.rc')
@with_application
def bucket_ls(application, project_id, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.bucket_array(
        application.query_service.find_buckets_in_first_kanban_view(project_id), is_json, is_jsonvja, custom_format)


# labels
@cli.group('label', help='Subcommand: label (see help)', aliases=['labels'])
def label_group():
    # vja label
    pass


@label_group.command('ls', help='Print labels ... (id; title)')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='Print as vja application json')
@click.option('custom_format', '--custom-format',
              help='Format with template from .vjacli/vja.rc')
@with_application
def label_ls(application, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.label_array(
        application.query_service.find_all_labels(), is_json, is_jsonvja, custom_format)


@label_group.command('add', help='Add label with title')
@click.argument('title', required=True, nargs=-1)
@with_application
def label_add(application, title):
    label = application.command_service.add_label(" ".join(title))
    click.echo(f'Created label {label.id}')


# tasks
@cli.command('add', aliases=['create'],
             help='Add new task')
@click.argument('title', required=True, nargs=-1)
@click.option('quiet_show', '-q', '--quiet-show', '--quiet', is_flag=True,
              help='Hide confirmation message')
@click.option('verbose_show', '-v', '--verbose-show', '--verbose', is_flag=True,
              help='Show resulting task when finished')
@click.option('project_id', '-o', '--project', '--project-id', '--project_id',
              help='Project (id or name), defaults to project from user settings, than to first favorite project')
@click.option('note', '-n', '--note', '--notes', '--description',
              help='Set description (note)')
@click.option('prio', '-p', '--prio', '--priority',
              help='Set priority')
@click.option('due', '-d', '--due', '--duedate', '--due-date', '--due_date',
              help='Set due date (supports parsedatetime expressions)')
@click.option('favorite', '-f', '--star', '--favorite', type=click.BOOL,
              help='Mark as favorite')
@click.option('label', '-l', '--label', multiple=True,
              help='Set label (label must exist on server)')
@click.option('reminder', '-r', '--alarm', '--remind', '--reminder', is_flag=False, flag_value='due',
              help='Set reminder (supports parsedatetime and timedelta expressions). '
                   'Absolute: "in 3 days at 18:00" or relative: "1h before due_date" or just -r to set equal to '
                   'due date.')
@click.option('force_create', '--force-create', '--force', is_flag=True,
              help='Force creation of non existing label')
@with_application
@click.pass_context
def task_add(ctx, application, title, quiet_show=False, verbose_show=False, **args):
    args_present = {k: v for k, v in args.items() if v is not None}
    task = application.command_service.add_task(" ".join(title), args_present.copy())
    if verbose_show or not quiet_show:
        click.echo(f'Created task {task.id} in project {task.project.id}')
    if verbose_show:
        ctx.invoke(task_show, tasks=[task.id])


@cli.command('clone', aliases=['copy'],
             help='Clone task with given task_id. Set the new title')
@click.argument('task_id', required=True, type=click.INT)
@click.argument('title', required=True, nargs=-1)
@click.option('quiet_show', '-q', '--quiet-show', '--quiet', is_flag=True,
              help='Hide confirmation message')
@click.option('verbose_show', '-v', '--verbose-show', '--verbose', is_flag=True,
              help='Show resulting task when finished')
@with_application
@click.pass_context
def task_clone(ctx, application, task_id, title, quiet_show=False, verbose_show=False):
    task = application.command_service.clone_task(task_id, " ".join(title))
    if verbose_show or not quiet_show:
        click.echo(f'Created task {task.id} in project {task.project.id} as clone from {task_id}')
    if verbose_show:
        ctx.invoke(task_show, tasks=[task.id])


@cli.command('edit', aliases=['modify', 'update'],
             help='Modify task/tasks. (Opens task in browser if no options are given)')
@click.argument('task_ids', required=True, type=click.INT, nargs=-1)
@click.option('quiet_show', '-q', '--quiet-show', '--quiet', is_flag=True,
              help='Hide confirmation message')
@click.option('verbose_show', '-v', '--verbose-show', '--verbose', is_flag=True,
              help='Show resulting task when finished')
@click.option('title', '-i', '--title',
              help='Set title')
@click.option('note', '-n', '--note', '--notes', '--description',
              help='Set description (note)')
@click.option('note_append', '-a', '--note-append', '--append-note', '--description-append', '--append-description',
              help='Append note to existing description separated by new line')
@click.option('prio', '-p', '--prio', '--priority', type=click.INT,
              help='Set priority')
@click.option('project_id', '-o', '--project', '--project-id', '--project_id',
              help='Move to project with id')
@click.option('position', '--project-position', '--project_position', '--position', type=click.INT,
              help='Set project position')
@click.option('bucket_id', '--bucket-id', '--bucket_id', type=click.INT,
              help='Set bucket id')
@click.option('kanban_position', '--kanban-position', '--kanban_position', type=click.INT,
              help='Set kanban position')
@click.option('due', '-d', '--due', '--duedate', '--due-date', '--due_date',
              help='Set due date (supports parsedatetime expressions)')
@click.option('favorite', '-f', '--favorite', '--star', type=click.BOOL,
              help='Mark as favorite')
@click.option('completed', '-c', '--completed', '--done', type=click.BOOL,
              help='Mark as completed')
@click.option('label', '-l', '--label',
              help='Set label (label must exist on server unless called with --force-create)')
@click.option('reminder', '-r', '--alarm', '--remind', '--reminder', is_flag=False, flag_value='due',
              help='Set reminder (supports parsedatetime and timedelta expressions). '
                   'Absolute: "in 3 days at 18:00" or relative: "1h30m before due_date" or just -r to set equal to '
                   'due date.')
@click.option('force_create', '--force-create', '--force', is_flag=True, default=None,
              help='Force creation of non existing label')
@with_application
@click.pass_context
def task_edit(ctx, application, task_ids, quiet_show=False, verbose_show=False, **args):
    args_present = {k: v for k, v in args.items() if v is not None}
    for task_id in task_ids:
        if args_present:
            task = application.command_service.edit_task(task_id, args_present.copy())
            if verbose_show or not quiet_show:
                click.echo(f'Modified task {task.id} in project {task.project.id}')
            if verbose_show:
                ctx.invoke(task_show, tasks=[task.id])
        else:
            application.open_browser_task(task_id)


@cli.command('toggle', aliases=['check', 'click', 'done'], help='Shortcut for marking / unmarking task as done')
@click.argument('task_id', required=True, type=click.INT)
@click.option('quiet_show', '-q', '--quiet-show', '--quiet', is_flag=True,
              help='Hide confirmation message')
@click.option('verbose_show', '-v', '--verbose-show', '--verbose', is_flag=True,
              help='Show resulting task when finished')
@with_application
@click.pass_context
def task_toggle(ctx, application, task_id, quiet_show=False, verbose_show=False):
    task = application.command_service.toggle_task_done(task_id)
    if verbose_show or not quiet_show:
        click.echo(f'Modified task {task.id} in project {task.project.id}')
    if verbose_show:
        ctx.invoke(task_show, tasks=[task_id])


@cli.command('defer', aliases=['delay'], help='Shortcut for moving the due_date and the reminders of the task. '
                                              'Examples for valid delay values are 2d, 1h30m.')
@click.argument('task_ids', required=True, type=click.INT, nargs=-1)
@click.argument('delay_by', required=True)
@click.option('quiet_show', '-q', '--quiet-show', '--quiet', is_flag=True,
              help='Hide confirmation message')
@click.option('verbose_show', '-v', '--verbose-show', '--verbose', is_flag=True,
              help='Show resulting task when finished')
@with_application
@click.pass_context
def task_defer(ctx, application, task_ids, delay_by, quiet_show=False, verbose_show=False):
    for task_id in task_ids:
        task = application.command_service.defer_task(task_id, delay_by)
        if verbose_show or not quiet_show:
            click.echo(f'Modified task {task.id} in project {task.project.id}')
        if verbose_show:
            ctx.invoke(task_show, tasks=[task_id])


@cli.command('ls', help='List tasks ... (task-id; priority; is_favorite; title; due_date; '
                        'has reminder; parent-project; project; labels; urgency). '
                        'Optionally limit output to given TASK_IDs.')
@click.argument('task_ids', type=click.INT, nargs=-1)
@click.option('is_json', '--json', default=False, is_flag=True,
              help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='Print as vja application json')
@click.option('custom_format', '--custom-format',
              help='Format with template from .vjacli/vja.rc')
@click.option('sort_string', '--sort', '--sort-by', '--sort-string',
              help='Sort by arguments (Arguments ar task attributes separated by ",") '
                   '(Prefix each criteria with "-" to reverse order')
@click.option('include_completed', '--include-completed', '--all', default=False, is_flag=True,
              help='Include completed tasks')
@click.option('due_date_filter', '-d', '--due', '--due-date', '--due_date',
              help='Filter by due date. The TEXT value must be like <operator> <value> '
                   'Shortcut for --filter="due_date <operator> <value>"')
@click.option('favorite_filter', '-f', '--favorite', '--star', type=click.BOOL,
              help='Filter by favorite flag. Shortcut for --filter="favorite_filter eq <value>"')
@click.option('general_filter', '--filter', multiple=True,
              help='General filter. Must be like <field> <operator> <value> e.g. --filter="priority ge 2" '
                   'where <operator> in (eq, ne, gt, lt, ge, le, before, after, contains). '
                   'Multiple occurrences of --filter are allowed and will be combined with logical AND.')
@click.option('label_filter', '-l', '--label',
              help='Filter by label (regex or id)')
@click.option('project_filter', '-o', '--project', '--project-id', '--project_id',
              help='Filter by project (regex or id)')
@click.option('upper_project_filter', '-t', '--base-project', '--top', '--base', '--upper-project',
              help='Filter by base project (regex or id). '
                   'All tasks whose project has the given argument as a ancestor are considered.')
@click.option('priority_filter', '-p', '--prio', '--priority',
              help='Filter by priority. The TEXT value must be like <operator> <value>, '
                   'Shortcut for --filter="priority <operator> <value>"')
@click.option('title_filter', '-i', '--title',
              help='Filter title (regex)')
@click.option('urgency_filter', '-u', '--urgent', '--urgency', is_flag=False, flag_value=3, type=click.INT,
              help='Filter by minimum urgency.  Shortcut for --filter="urgency ge <value>"')
@with_application
def task_ls(application, task_ids, is_json, is_jsonvja, custom_format, include_completed, sort_string=None,
            **filter_args):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    filter_args = {k: v for k, v in filter_args.items() if v is not None}

    tasks = application.query_service.find_filtered_tasks(include_completed, sort_string, filter_args)
    if task_ids:
        tasks = [t for t in tasks if t.id in task_ids]

    application.output.task_array(tasks, is_json, is_jsonvja, custom_format)
    if not is_json and not is_jsonvja and not custom_format:
        click.echo(f"Count: {len(list(tasks))}")


@cli.command('show', help='Show task details. Multiple task ids may be given')
@click.argument('tasks', type=click.INT, nargs=-1)
@click.option('is_json', '--json', default=False, is_flag=True,
              help='Print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='Print as vja application json')
@with_application
def task_show(application, tasks, is_json, is_jsonvja):
    for task_id in tasks:
        task = application.query_service.find_task_by_id(task_id)
        application.output.task(task, is_json, is_jsonvja)


@cli.command(name='open', help='Open task in browser. Multiple task ids may be given. '
                               'If tasks is empty, then open vikunjas starting page.')
@click.argument('tasks', required=False, type=click.INT, nargs=-1)
@with_application
def task_open(application, tasks):
    if not tasks:
        application.open_browser_task('')
    else:
        for task_id in tasks:
            application.open_browser_task(task_id)


# task group
@cli.group('task', help='(optional) subcommand: task (see help)', aliases=['tasks'])
def task_group():
    # vja task
    pass


task_group.add_command(task_add)
task_group.add_command(task_clone)
task_group.add_command(task_edit)
task_group.add_command(task_toggle)
task_group.add_command(task_defer)
task_group.add_command(task_ls)
task_group.add_command(task_show)
task_group.add_command(task_open)


@cli.command('logout', help='Remove local access token')
@with_application
def logout(application):
    application.command_service.logout()


if __name__ == '__main__':
    cli()



import dataclasses
import typing
from dataclasses import dataclass, field
from datetime import datetime, timedelta

from vja.parse import html2text, parse_json_date

ID_TITLE = 'id={},title={}'


def custom_output(cls):
    hidden_attribute_names = ['json', 'description_text']

    def __str__(self):
        """Returns a string containing only the non-null attribute values, excluding json attribute ."""
        return '\n'.join(f'{attribute.name}: {_str_value(getattr(self, attribute.name))}'
                         for attribute in dataclasses.fields(self)
                         if attribute.name not in hidden_attribute_names and getattr(self, attribute.name))

    def _str_value(v):
        if isinstance(v, datetime):
            return v.isoformat()
        if isinstance(v, (Project, ProjectView, Label, TaskReminder)):
            return v.short_str()
        if isinstance(v, list):
            return [_str_value(x) for x in v]
        return str(v)

    setattr(cls, '__str__', __str__)
    return cls


def data_dict(cls):
    def data_dict_function(self):
        return {k: _transform_value(v) for k, v in self.__dict__.items() if k != 'json'}

    def _transform_value(v):
        if isinstance(v, datetime):
            return v.isoformat()
        if _is_data_dict(v):
            return v.data_dict()
        if isinstance(v, list):
            return [_transform_value(x) for x in v]
        return v

    def _is_data_dict(v):
        return hasattr(v, 'data_dict') and callable(v.data_dict)

    setattr(cls, 'data_dict', data_dict_function)
    return cls


@dataclass(frozen=True)
@data_dict
class User:
    json: dict = field(repr=False)
    id: int
    username: str
    name: str
    default_project_id: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['username'], json['name'], json['settings']['default_project_id'])


@dataclass
@data_dict
# pylint: disable=too-many-instance-attributes
class ProjectView:
    json: dict = field(repr=False)
    id: int
    title: str
    project_id: int
    view_kind: str  # The kind of this view. Can be list, gantt, table or kanban.
    bucket_configuration_mode: str  # Can be none, manual or filter. manual
    default_bucket_id: int
    done_bucket_id: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'],
                   json['project_id'],
                   json['view_kind'],
                   json['bucket_configuration_mode'],
                   json['default_bucket_id'],
                   json['done_bucket_id'])

    @classmethod
    def from_json_array(cls, json_array):
        return [ProjectView.from_json(x) for x in json_array or []]

    def short_str(self):
        return ID_TITLE.format(self.id, self.title)


@dataclass
@custom_output
@data_dict
# pylint: disable=too-many-instance-attributes
class Project:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    is_favorite: bool
    is_archived: bool
    parent_project_id: int
    ancestor_projects: typing.List['Project']
    views: typing.List[ProjectView]

    @classmethod
    def from_json(cls, json, ancestor_projects):
        return cls(json, json['id'], json['title'], json['description'],
                   json['is_favorite'], json['is_archived'],
                   json['parent_project_id'],
                   ancestor_projects,
                   ProjectView.from_json_array(json['views']))

    @classmethod
    def from_json_array(cls, json_array, ancestor_projects):
        return [Project.from_json(x, ancestor_projects) for x in json_array or []]

    def get_first_kanban_project_view(self):
        return next(x for x in self.views if x.view_kind == "kanban")

    def short_str(self):
        return ID_TITLE.format(self.id, self.title)


@dataclass(frozen=True)
@data_dict
class Bucket:
    json: dict = field(repr=False)
    id: int
    title: str
    limit: int
    position: int
    count_tasks: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'],
                   json['limit'],
                   json['position'],
                   json['count'])

    @classmethod
    def from_json_array(cls, json_array):
        return [Bucket.from_json(x) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class Label:
    json: dict = field(repr=False)
    id: int
    title: str

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'])

    @classmethod
    def from_json_array(cls, json_array):
        return [Label.from_json(x) for x in json_array or []]

    def short_str(self):
        return ID_TITLE.format(self.id, self.title)


@dataclass
@custom_output
@data_dict
# pylint: disable=too-many-instance-attributes
class TaskReminder:
    json: dict = field(repr=False)
    reminder: datetime
    relative_period: int
    relative_to: str

    @classmethod
    def from_json(cls, json):
        return cls(json, parse_json_date(json['reminder']), json['relative_period'], json['relative_to'])

    @classmethod
    def from_json_array(cls, json_array):
        return [TaskReminder.from_json(x) for x in json_array or []]

    def short_str(self):
        return 'reminder={},period={},relative_to={}'.format(
            self.reminder.isoformat() if self.reminder else ' ',
            self.relative_period,
            self.relative_to)


@dataclass
@custom_output
@data_dict
# pylint: disable=too-many-instance-attributes
class Task:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    description_text: str
    priority: int
    is_favorite: bool
    due_date: datetime
    reminders: typing.List[TaskReminder]
    repeat_mode: int
    repeat_after: timedelta
    start_date: datetime
    end_date: datetime
    percent_done: float
    done: bool
    done_at: datetime
    label_objects: typing.List[Label]
    project: Project
    position: int
    bucket_id: int
    created: datetime
    updated: datetime
    urgency: float = field(init=False)

    @property
    def labels(self):
        return ",".join(map(lambda label: label.title, self.label_objects or []))

    @classmethod
    def from_json(cls, json, project_object, labels):
        return cls(json, json['id'], json['title'],
                   json['description'],
                   html2text(json['description']),
                   json['priority'],
                   json['is_favorite'],
                   parse_json_date(json['due_date']),
                   TaskReminder.from_json_array(json["reminders"]),
                   json['repeat_mode'],
                   timedelta(seconds=json['repeat_after']),
                   parse_json_date(json['start_date']),
                   parse_json_date(json['end_date']),
                   json['percent_done'],
                   json['done'],
                   parse_json_date(json['done_at']),
                   labels,
                   project_object,
                   json['position'],
                   json['bucket_id'],
                   parse_json_date(json['created']),
                   parse_json_date(json['updated'])
                   )

    def has_label(self, label):
        return any(x.id == label.id for x in self.label_objects)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

requirements = [
    "click>=8.1",
    "click-aliases>=1",
    "requests>=2.31",
    "parsedatetime>=2.6",
    "python-dateutil>=2.8"
]

setup(
    name="vja",
    author="ce72",
    description="A simple CLI for Vikunja task manager",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/ce72/vja",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    entry_points={'console_scripts': [
        'vja=vja.cli:cli'
    ]},
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)



import logging
import operator
import re
import typing
from datetime import datetime, timedelta

from vja.parse import (parse_bool_arg, parse_date_arg_to_datetime,
                       parse_date_arg_to_timedelta, rgetattr)

logger = logging.getLogger(__name__)


def _create_due_date_filter(value: str):
    if value.strip() == '':
        return lambda x: not x.due_date
    return _create_general_filter([f'due_date {value}'])


def _create_general_filter(values: typing.List[str]):
    return [_create_single_general_filter(x) for x in values]


def _create_single_general_filter(value: str):
    arguments = value.split(" ", 2)
    field = arguments[0]
    operation = _operators[arguments[1]]
    value = arguments[2]
    logger.debug("general filter %s: %s %s", field, operation.__name__, value)
    return lambda x: _general_filter(x, field, operation, value)


def _general_filter(task, field_name, operation, value):
    task_value = rgetattr(task, field_name)
    if task_value is not None:
        if isinstance(task_value, datetime):
            value = parse_date_arg_to_datetime(value)
        elif isinstance(task_value, timedelta):
            value = parse_date_arg_to_timedelta(value)
        elif isinstance(task_value, bool):
            value = parse_bool_arg(value)
        elif isinstance(task_value, int):
            value = int(value)
        elif isinstance(task_value, float):
            value = float(value)
        return operation(task_value, value)
    return False


def _create_favorite_filter(value):
    return _create_general_filter([f'is_favorite eq {value}'])


def _create_label_filter(value):
    logger.debug("filter labels %s", value)
    if str(value).isdigit():
        return lambda x: any(label.id == int(value) for label in x.label_objects)
    if str(value).strip() == '':
        return lambda x: not x.label_objects
    return lambda x: any(_find_regex(value, label.title) for label in x.label_objects)


def _create_project_filter(value):
    if str(value).isdigit():
        return _create_general_filter([f'project.id eq {value}'])
    return lambda task: _find_regex(value, task.project.title)


def _create_upper_project_filter(value):
    if str(value).isdigit():
        return lambda x: x.project.id == int(value) \
            or any(ancestor.id == int(value) for ancestor in x.project.ancestor_projects)
    return lambda task: _find_regex(value, task.project.title) \
        or any(_find_regex(value, ancestor.title) for ancestor in task.project.ancestor_projects)


def _create_title_filter(value):
    return lambda task: _find_regex(value, task.title)


def _find_regex(regex, value):
    return bool(re.search(re.compile(regex, re.IGNORECASE), value))


def _create_priority_filter(value):
    return _create_general_filter([f'priority {value}'])


def _create_urgency_filter(value):
    return _create_general_filter([f'urgency ge {value}'])


_operators = {
    'eq': operator.eq,
    'ne': operator.ne,
    'lt': operator.lt,
    'le': operator.le,
    'gt': operator.gt,
    'ge': operator.ge,
    'before': operator.lt,
    'after': operator.gt,
    'contains': operator.contains
}

_filter_mapping = {
    'due_date_filter': _create_due_date_filter,
    'favorite_filter': _create_favorite_filter,
    'general_filter': _create_general_filter,
    'label_filter': _create_label_filter,
    'project_filter': _create_project_filter,
    'upper_project_filter': _create_upper_project_filter,
    'title_filter': _create_title_filter,
    'priority_filter': _create_priority_filter,
    'urgency_filter': _create_urgency_filter,
}


def create_filters(filter_args):
    filters = []
    for filter_name, filter_value in filter_args.items():
        add_filter = _filter_mapping[filter_name](filter_value)
        filters.extend(add_filter if isinstance(add_filter, (list, tuple)) else [add_filter])
    return filters
